// file SharedMemoryMapGuard.h
// author Edda Nourbakhsh


#pragma once


#include "SharedMemoryGuard.h"


  template < class TMapPtr > class SharedMemoryMapGuard
    /// allows different DLLs of the same process implementing the same type 
    /// of std::map to register their map in the map stored in this segment of 
    /// boost::interprocess::managed_shared_memory.
    /// <b>NOTE:</b> SharedMemoryMapGuard is for 
    /// sharing memory between DLLs; it <b>DOES NOT</b> work between processes.
  // ************************************************************************
  {


  public:
    typedef std::map< void *, TMapPtr > mMapType;

  private:
    typedef void * segmentOwnerType;
    typedef mMapType * mMapPtrType;


  public:
    SharedMemoryMapGuard(
      char const * memoryName,
      char const * segmentName,
      TMapPtr ptr = nullptr
      )
      :
      mPtr(nullptr),
      mSegmentName(segmentName),
      mSegmentOwnerName(SegmentOwnerAppendix),
      mSegmentOwner(nullptr),
      mSharedMemoryGuard(memoryName)
      /// Default constructor
      /// instantiates the SharedMemoryGuard object managing 
      /// boost::interprocess::managed_shared_memory.
      /// If the indicated shared memory segment does not exist yet, it 
      /// constructs the segment, registers as segment owner, and instantiates 
      /// the segment map.\n
      /// In any case it inserts the parameter ptr into the map of this segment.
      ///
      /// \param [in] memoryName Name of 
      ///                         boost::interprocess::managed_shared_memory
      /// \param [in] segmentName Name of the shared memory segment
      /// \param [in] ptr         Pointer to insert in the map of this segment
      //-----------------------------------------------------------------------
    {
      mSegmentOwnerName.append(mSegmentName);


      //Find the shared pointer to mSegmentOwner.
      std::pair<
        segmentOwnerType *,
        boost::interprocess::managed_shared_memory::size_type
      > findSegmentOwner
      = mSharedMemoryGuard.mMemory.find < segmentOwnerType >
      (mSegmentOwnerName.c_str()); //(object name)

      if (findSegmentOwner.second >= 1)
      {
        mSegmentOwner = findSegmentOwner.first;
      }
      else
      {
        //Construct the shared pointer to mSegmentOwner
        mSegmentOwner = mSharedMemoryGuard.mMemory.construct <
          segmentOwnerType
        >
        (mSegmentOwnerName.c_str()) //(object name)
        (); // (ctor parameter)

        *mSegmentOwner = this;
      }

      //Find the shared map pointer
      std::pair<
        mMapPtrType *,
        boost::interprocess::managed_shared_memory::size_type
      > findSharedSegmentResult
      = mSharedMemoryGuard.mMemory.find< mMapPtrType >
      (mSegmentName.c_str()); //(object name)

      if (findSharedSegmentResult.second >= 1)
      {
        mPtr = findSharedSegmentResult.first;
      }
      else
      {
        //Construct the shared map pointer
        mPtr = mSharedMemoryGuard.mMemory.construct <
          mMapPtrType
        >
        (mSegmentName.c_str()) //(object name)
        (); // (ctor parameters)

        *mPtr = &mMapInstance;
      }

      if (mPtr != nullptr)
      {
        if (*mPtr != nullptr)
        {
          (*mPtr)->insert(mMapType::value_type((void *)this, ptr));
        }
      }
    }

    ///////////////

    ~SharedMemoryMapGuard()
      /// Default destructor 
      /// removes the shared memory segment if it is the segment owner.
      //-----------------------------------------------------------------------
    {
      if (mSharedMemoryGuard.mHasSharedMemory && *mSegmentOwner == this)
      {
        *mPtr = nullptr;
        mPtr = nullptr;

        //We're done, delete all the objects
        bool destroyed = mSharedMemoryGuard.mMemory.destroy<
          mMapPtrType
        >(mSegmentName.c_str());

        *mSegmentOwner = nullptr;
        mSegmentOwner = nullptr;

        //We're done, delete all the objects
        mSharedMemoryGuard.mMemory.destroy<
          segmentOwnerType
        >(mSegmentOwnerName.c_str());
      }
    }

    ///////////////

  public:
    mMapType * getMapPtr() const
      ///
      /// \returns pointer to the map, which is shared by this shared memory 
      /// segment.
      //-----------------------------------------------------------------------
    {
      return *mPtr;
    }

    ///////////////

  private:
    boost::interprocess::string mSegmentName;

    /// SharedMemoryGuard object managing
    /// boost::interprocess::managed_shared_memory.
    SharedMemoryGuard mSharedMemoryGuard;

    /// Pointer to the map, which is shared by this shared memory segment
    mMapPtrType * mPtr;

    /// If this SharedMemoryMapGuard object is the creator of the shared 
    /// memory segment, this map instance is shared by this segment.
    /// Otherwise this map instance won't be used.
    mMapType mMapInstance;

    boost::interprocess::string mSegmentOwnerName;
    segmentOwnerType * mSegmentOwner;


  };
