// file SharedMemoryPtrGuard.h
// author Edda Nourbakhsh


#pragma once


#include "SharedMemoryGuard.h"


  template < class TPtrTarget > class SharedMemoryPtrGuard
    /// manages a segment of boost::interprocess::managed_shared_memory storing
    /// pointer to TPtrTarget. <b>NOTE:</b> SharedMemoryPtrGuard is for 
    /// sharing memory between DLLs; it <b>DOES NOT</b> work between processes.
    /// SharedMemoryPtrGuard provides the pointer stored in the indicated
    /// shared memory segment. Only if the segment does not exist yet, 
    /// SharedMemoryPtrGuard creates the segment and assigns the 
    /// constructor parameter ptr to the shared memory pointer.
  {


  public:
    SharedMemoryPtrGuard(
      char const * memoryName,
      char const * segmentName,
      TPtrTarget ptr = nullptr
      )
      :
      mPtr(nullptr),
      mSegmentName(segmentName),
      mSegmentOwnerName(SegmentOwnerAppendix),
      mSegmentOwner(nullptr),
      mSharedMemoryGuard(memoryName)
      /// Default constructor
      /// instantiates the SharedMemoryGuard object managing 
      /// boost::interprocess::managed_shared_memory.
      /// If it finds the indicated shared memory segment, it sets the .
      /// member mPtr to the found address.
      /// Otherwise it constructs the segment, registers as segment owner, and 
      /// assigns the parameter ptr to the shared memory pointer.
      ///
      /// \param [in] memoryName Name of 
      ///                         boost::interprocess::managed_shared_memory
      /// \param [in] segmentName Name of the shared memory segment
      /// \param [in] ptr         Pointer to keep in shared memory (unless it 
      ///                         is already assigned)
    {
      mSegmentOwnerName.append(mSegmentName);

      //Find the shared pointer to mSegmentOwner.
      std::pair<
        void * *,
        boost::interprocess::managed_shared_memory::size_type
      > findSegmentOwner
      = mSharedMemoryGuard.mMemory.find < void * >
      (mSegmentOwnerName.c_str()); //(object name)

      if (findSegmentOwner.second >= 1)
      {
        mSegmentOwner = findSegmentOwner.first;
      }
      else
      {
        //Construct the shared pointer to mSegmentOwner
        mSegmentOwner = mSharedMemoryGuard.mMemory.construct <
          void *
        >
        (mSegmentOwnerName.c_str()) //(object name)
        (); // (ctor parameter)

        *mSegmentOwner = this;
      }

      //Find the shared pointer
      std::pair<
        TPtrTarget *,
        boost::interprocess::managed_shared_memory::size_type
      > findSharedSegmentResult
      = mSharedMemoryGuard.mMemory.find< TPtrTarget >
      (mSegmentName.c_str()); //(object name)

      if (findSharedSegmentResult.second >= 1)
      {
        mPtr = findSharedSegmentResult.first;
      }
      else
      {
        //Construct the shared pointer
        mPtr = mSharedMemoryGuard.mMemory.construct <
          TPtrTarget
        >
        (mSegmentName.c_str()) //(object name)
        (); // (ctor parameter)

        *mPtr = ptr;
      }
    }

    ///////////////

    ~SharedMemoryPtrGuard()
      /// Default destructor 
      /// removes the shared memory segment if it is the segment owner.
    {
      if (mSharedMemoryGuard.mHasSharedMemory && *mSegmentOwner == this)
      {
        *mPtr = nullptr;
        mPtr = nullptr;

        bool destroyed = mSharedMemoryGuard.mMemory.destroy<
          TPtrTarget
        >(mSegmentName.c_str());

        *mSegmentOwner = nullptr;
        mSegmentOwner = nullptr;

        mSharedMemoryGuard.mMemory.destroy<
          void *
        >(mSegmentOwnerName.c_str());
      }
    }

    ///////////////

  public:
    TPtrTarget * get() const
      ///
      /// \returns pointer stored in this shared memory segment.
    {
      return mPtr;
    }


  private:
    boost::interprocess::string mSegmentName;

    /// SharedMemoryGuard object managing
    /// boost::interprocess::managed_shared_memory.
    SharedMemoryGuard mSharedMemoryGuard;

    /// Pointer stored in this shared memory segment
    TPtrTarget * mPtr;

    boost::interprocess::string mSegmentOwnerName;
    void * * mSegmentOwner;


  };
